// This file is part of the uutils coreutils package.
//
// For the full copyright and license information, please view the LICENSE
// file that was distributed with this source code.

uu_checksum_common::declare_standalone!("sha256sum", uucore::checksum::AlgoKind::Sha256);
