<style>
    dd {
        margin-bottom: 1em;
    }
    main {
        position: relative;
    }
    .additional {
        position: absolute;
        top: 0.5em;
        right: 0;
        display: flex;
        gap: 5px;
        align-items: center;
        font-size: 1.3em;
    }
    .platforms {
        font-size: 2rem;
    }
    dd > p {
        margin-top: 0.2em;
    }
</style>
